from __future__ import annotations

import importlib.metadata
from pathlib import Path
from typing import Optional

import typer
from deciphon_schema import HMMFile
from typer import echo

from h3daemon.daemon import Daemon
from h3daemon.daemonize import spawn
from h3daemon.pidfile import create_pidfile

__all__ = ["app"]


app = typer.Typer(
    add_completion=False,
    pretty_exceptions_short=True,
    pretty_exceptions_show_locals=False,
)

O_VERSION = typer.Option(None, "--version", is_eager=True)
O_PORT = typer.Option(0, help="Port to listen to.")
O_FORCE = typer.Option(False, "--force")
O_WAIT = typer.Option(False, "--wait")
O_STDIN = typer.Option(None, "--stdin")
O_STDOUT = typer.Option(None, "--stdout")
O_STDERR = typer.Option(None, "--stderr")


@app.callback(invoke_without_command=True)
def cli(version: Optional[bool] = O_VERSION):
    if version:
        echo(importlib.metadata.version(str(__package__)))
        raise typer.Exit()


@app.command()
def start(
    hmmfile: Path,
    port: int = O_PORT,
    stdin: Optional[Path] = O_STDIN,
    stdout: Optional[Path] = O_STDOUT,
    stderr: Optional[Path] = O_STDERR,
    force: bool = O_FORCE,
):
    """
    Start daemon.
    """
    hmm = HMMFile(path=hmmfile)
    spawn(
        hmm,
        cport=port,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        force=force,
    )


@app.command()
def stop(hmmfile: Path, force: bool = O_FORCE):
    """
    Stop daemon.
    """
    hmm = HMMFile(path=hmmfile)
    pidfile = create_pidfile(hmm.path)
    x = Daemon.possess(pidfile)
    x.shutdown(force=force)


@app.command()
def ready(hmmfile: Path, wait: bool = O_WAIT):
    """
    Check if h3daemon is running and ready.
    """
    file = HMMFile(path=hmmfile)
    pidfile = create_pidfile(file.path)
    x = Daemon.possess(pidfile)
    echo(x.healthy())



from __future__ import annotations

import hmmer
import psutil

from h3daemon.polling import Polling
from h3daemon.tcp import tcp_connections

__all__ = ["Worker"]


class Worker:
    def __init__(self, proc: psutil.Process):
        self._proc = proc

    @property
    def process(self):
        return self._proc

    @staticmethod
    def cmd(wport: int):
        hmmpgmd = hmmer.path(hmmer.hmmpgmd)
        return [hmmpgmd, "--worker", "127.0.0.1", "--cpu", "1", "--wport", str(wport)]

    def healthy(self):
        if not self._proc.is_running():
            return False
        try:
            lports = self.local_established_ports()
            rports = self.remote_established_ports()
        except psutil.ZombieProcess:
            # psutil bug: https://github.com/giampaolo/psutil/issues/2116
            return True
        return len(lports) > 0 and len(rports) > 0

    def wait_for_readiness(self):
        for attempt in Polling():
            with attempt:
                assert self.healthy()

    def local_established_ports(self):
        connections = tcp_connections(self._proc)
        connections = [x for x in connections if x.status == "ESTABLISHED"]
        return [x.laddr.port for x in connections if x.laddr.ip == "127.0.0.1"]

    def remote_established_ports(self):
        connections = tcp_connections(self._proc)
        connections = [x for x in connections if x.status == "ESTABLISHED"]
        return [x.raddr.port for x in connections if x.raddr.ip == "127.0.0.1"]



from h3daemon.master import Master
from h3daemon.worker import Worker


def assert_peers_healthy(master: Master, worker: Worker):
    assert master.healthy()
    assert worker.healthy()
    master_listen = master.local_listening_ports()
    master_lport = master.local_established_ports()
    master_rport = master.remote_established_ports()
    worker_lport = worker.local_established_ports()
    worker_rport = worker.remote_established_ports()

    assert len(master_lport) >= 1
    assert len(worker_rport) == 1
    assert worker_rport[0] in master_lport
    assert len(master_rport) >= 1
    assert len(worker_lport) == 1
    assert worker_lport[0] in master_rport
    assert len(master_listen) == 2
    master_ports = set(master_listen)
    assert len(master_ports) == 2
    master_ports.remove(worker_rport[0])
    assert len(master_ports) == 1



from h3daemon.daemon import Daemon, daemon_context
from h3daemon.daemonize import spawn
from h3daemon.ensure_pressed import ensure_pressed
from h3daemon.possess import possess

__all__ = ["Daemon", "ensure_pressed", "spawn", "possess", "daemon_context"]



from pathlib import Path

from pidlockfile import PIDLockFile

__all__ = ["create_pidfile"]


def create_pidfile(file: Path, timeout=5):
    return PIDLockFile(f"{str(file.absolute())}.pid", timeout=timeout)



from pidlockfile import PIDLockFile
from h3daemon.polling import polling

from h3daemon.daemon import Daemon

__all__ = ["possess"]


@polling
def possess(pidfile: PIDLockFile):
    return Daemon.possess(pidfile)



from multiprocessing import Process
from typing import Any, Optional

from daemon import DaemonContext
from deciphon_schema import HMMFile

from h3daemon.daemon import Daemon
from h3daemon.ensure_pressed import ensure_pressed
from h3daemon.errors import DaemonAlreadyRunningError
from h3daemon.pidfile import create_pidfile

__all__ = ["daemonize", "spawn"]


def daemonize(
    hmmfile: HMMFile,
    cport: int = 0,
    wport: int = 0,
    stdin: Optional[Any] = None,
    stdout: Optional[Any] = None,
    stderr: Optional[Any] = None,
    detach: Optional[bool] = None,
):
    fin = open(stdin, "r") if stdin else stdin
    fout = open(stdout, "w+") if stdout else stdout
    ferr = open(stderr, "w+") if stderr else stderr

    pidfile = create_pidfile(hmmfile.path)
    assert pidfile.is_locked() is None
    with DaemonContext(
        working_directory=str(hmmfile.path.parent),
        pidfile=pidfile,
        detach_process=detach,
        stdin=fin,
        stdout=fout,
        stderr=ferr,
    ):
        x = Daemon.spawn(hmmfile, cport, wport)
        x.join()

    return pidfile


def spawn(
    hmmfile: HMMFile,
    cport: int = 0,
    wport: int = 0,
    stdin: Optional[Any] = None,
    stdout: Optional[Any] = None,
    stderr: Optional[Any] = None,
    detach: Optional[bool] = None,
    force: Optional[bool] = False,
):
    ensure_pressed(hmmfile)
    pidfile = create_pidfile(hmmfile.path)
    if pidfile.is_locked():
        if not force:
            raise DaemonAlreadyRunningError(f"Daemon for {hmmfile} is already running.")
        x = Daemon.possess(pidfile)
        x.shutdown(force=force)

    args = (hmmfile, cport, wport, stdin, stdout, stderr, detach)
    p = Process(target=daemonize, args=args, daemon=False)
    p.start()
    return pidfile



from pathlib import Path
from subprocess import check_call

import hmmer
from deciphon_schema import HMMFile

__all__ = ["ensure_pressed"]

pressed_extensions = ["h3f", "h3i", "h3m", "h3p"]


def ensure_pressed(hmmfile: HMMFile):
    for x in pressed_extensions:
        filename = Path(f"{hmmfile.path}.{x}")
        if not filename.exists():
            for x in pressed_extensions:
                Path(f"{hmmfile.path}.{x}").unlink(True)
            check_call([hmmer.path(hmmer.hmmpress), str(hmmfile.path)])



import os
import platform
import traceback
from contextlib import contextmanager, suppress

import psutil
from deciphon_schema import HMMFile
from pidlockfile import PIDLockFile

from h3daemon.ensure_pressed import ensure_pressed
from h3daemon.errors import ChildNotFoundError, PIDNotFoundError
from h3daemon.healthy import assert_peers_healthy
from h3daemon.master import Master
from h3daemon.port import find_free_port
from h3daemon.worker import Worker

__all__ = ["Daemon", "daemon_context"]


def shutdown(x: psutil.Process, force: bool):
    with suppress(psutil.NoSuchProcess):
        if force:
            x.kill()
            x.wait()
        else:
            x.terminate()
            try:
                x.wait(3)
            except psutil.TimeoutExpired:
                shutdown(x, True)


def debug_exception(exception: Exception):
    if os.environ.get("H3DAEMON_DEBUG", 0):
        with open("h3daemon_debug.txt", "a") as f:
            f.write(f"{type(exception)}\n")
            f.write(f"{exception.args}\n")
            f.write(traceback.format_exc())


def debug_msg(msg: str):
    if os.environ.get("H3DAEMON_DEBUG", 0):
        with open("h3daemon_debug.txt", "a") as f:
            f.write(f"{msg}\n")


class Daemon:
    def __init__(self, master: Master, worker: Worker, process: psutil.Process | None):
        self._master = master
        self._worker = worker
        self._process = process

    @classmethod
    def spawn(cls, hmmfile: HMMFile, cport: int = 0, wport: int = 0):
        master: Master | None = None
        worker: Worker | None = None

        try:
            cport = find_free_port() if cport == 0 else cport
            wport = find_free_port() if wport == 0 else wport

            cmd = Master.cmd(cport, wport, str(hmmfile.path))
            master = Master(psutil.Popen(cmd))
            master.wait_for_readiness()

            cmd = Worker.cmd(wport)
            worker = Worker(psutil.Popen(cmd))
            worker.wait_for_readiness()

            if platform.system() != "Darwin":
                assert_peers_healthy(master, worker)
        except Exception as exception:
            debug_exception(exception)
            if worker:
                shutdown(worker.process, force=True)
            if master:
                shutdown(master.process, force=True)
            raise exception

        return cls(master, worker, None)

    @classmethod
    def possess(cls, pidfile: PIDLockFile):
        pid = pidfile.is_locked()
        if not pid:
            raise PIDNotFoundError("PID not in pidfile.")
        process = psutil.Process(pid)
        children = process.children()

        if len(children) > 0:
            master = Master(children[0])
        else:
            raise ChildNotFoundError("Master not found.")

        if len(children) > 1:
            worker = Worker(children[1])
        else:
            raise ChildNotFoundError("Worker not found.")

        return cls(master, worker, process)

    def shutdown(self, force=False):
        if self._worker is not None:
            shutdown(self._worker.process, True)

        if self._master is not None:
            shutdown(self._master.process, True)

        if self._process is not None:
            shutdown(self._process, force)

    def healthy(self) -> bool:
        try:
            if self._process:
                assert self._process.is_running()
            assert_peers_healthy(self._master, self._worker)
        except Exception as exception:
            debug_exception(exception)
            return False
        return True

    def port(self) -> int:
        master_ports = set(self._master.local_listening_ports())
        master_ports.remove(self._worker.remote_established_ports()[0])
        return int(list(master_ports)[0])

    def join(self):
        psutil.wait_procs([self._master.process, self._worker.process])


@contextmanager
def daemon_context(hmmfile: HMMFile, cport: int = 0, wport: int = 0):
    ensure_pressed(hmmfile)
    x = Daemon.spawn(hmmfile, cport, wport)
    try:
        yield x
    finally:
        x.shutdown()



import functools

from tenacity import Retrying, retry, stop_after_delay, wait_fixed, wait_random

STOP = stop_after_delay(10)
WAIT = wait_fixed(0.1) + wait_random(0, 0.1)


def polling(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        cb = retry(
            func,
            stop=STOP,
            wait=WAIT,
        )
        return cb(*args, **kwargs)

    return wrapper


def Polling():
    return Retrying(stop=STOP, wait=WAIT)



__all__ = ["ChildNotFoundError", "DaemonAlreadyRunningError", "PIDNotFoundError"]


class ChildNotFoundError(RuntimeError):
    def __str__(self):
        return repr(self)


class DaemonAlreadyRunningError(RuntimeError):
    def __str__(self):
        return repr(self)


class PIDNotFoundError(RuntimeError):
    def __str__(self):
        return repr(self)



from psutil import Process

__all__ = ["tcp_connections"]


def tcp_connections(x: Process):
    # psutil bug: https://github.com/giampaolo/psutil/issues/2116
    with open("/dev/null", "wb"):
        connections = x.net_connections(kind="tcp")
    return connections
